# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass, field
from typing import Any, Dict, List
from collections.abc import Iterable
import zaber_bson
from ..ascii.trigger_enabled_state import TriggerEnabledState


@dataclass
class TriggerEnabledStates:

    states: List[TriggerEnabledState] = field(default_factory=list)

    @staticmethod
    def zero_values() -> 'TriggerEnabledStates':
        return TriggerEnabledStates(
            states=[],
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'TriggerEnabledStates':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return TriggerEnabledStates.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'states': [item.to_dict() for item in self.states] if self.states is not None else [],
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'TriggerEnabledStates':
        return TriggerEnabledStates(
            states=[TriggerEnabledState.from_dict(item) for item in data.get('states')],  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.states is not None:
            if not isinstance(self.states, Iterable):
                raise ValueError('Property "States" of "TriggerEnabledStates" is not iterable.')

            for i, states_item in enumerate(self.states):
                if states_item is None:
                    raise ValueError(f'Item {i} in property "States" of "TriggerEnabledStates" is None.')

                if not isinstance(states_item, TriggerEnabledState):
                    raise ValueError(f'Item {i} in property "States" of "TriggerEnabledStates" is not an instance of "TriggerEnabledState".')

                states_item.validate()



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #
from typing import TYPE_CHECKING, List
from ..dto import requests as dto
from ..units import Units, LengthUnits, VelocityUnits, AccelerationUnits, TimeUnits
from ..call import call, call_async, call_sync
from ..dto.ascii.lockstep_axes import LockstepAxes

if TYPE_CHECKING:
    from .device import Device


class Lockstep:
    """
    Represents a lockstep group with this ID on a device.
    A lockstep group is a movement synchronized pair of axes on a device.
    Requires at least Firmware 6.15 or 7.11.
    """

    @property
    def device(self) -> 'Device':
        """
        Device that controls this lockstep group.
        """
        return self._device

    @property
    def lockstep_group_id(self) -> int:
        """
        The number that identifies the lockstep group on the device.
        """
        return self._lockstep_group_id

    def __init__(self, device: 'Device', lockstep_group_id: int):
        self._device = device
        self._lockstep_group_id = lockstep_group_id

    def enable(
            self,
            *axes: int
    ) -> None:
        """
        Activate the lockstep group on the axes specified.

        Args:
            axes: The numbers of axes in the lockstep group.
        """
        request = dto.LockstepEnableRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
            axes=list(axes),
        )
        call("device/lockstep_enable", request)

    async def enable_async(
            self,
            *axes: int
    ) -> None:
        """
        Activate the lockstep group on the axes specified.

        Args:
            axes: The numbers of axes in the lockstep group.
        """
        request = dto.LockstepEnableRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
            axes=list(axes),
        )
        await call_async("device/lockstep_enable", request)

    def disable(
            self
    ) -> None:
        """
        Disable the lockstep group.
        """
        request = dto.LockstepDisableRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
        )
        call("device/lockstep_disable", request)

    async def disable_async(
            self
    ) -> None:
        """
        Disable the lockstep group.
        """
        request = dto.LockstepDisableRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
        )
        await call_async("device/lockstep_disable", request)

    def stop(
            self,
            wait_until_idle: bool = True
    ) -> None:
        """
        Stops ongoing lockstep group movement. Decelerates until zero speed.

        Args:
            wait_until_idle: Determines whether function should return after the movement is finished or just started.
        """
        request = dto.LockstepStopRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
            wait_until_idle=wait_until_idle,
        )
        call("device/lockstep_stop", request)

    async def stop_async(
            self,
            wait_until_idle: bool = True
    ) -> None:
        """
        Stops ongoing lockstep group movement. Decelerates until zero speed.

        Args:
            wait_until_idle: Determines whether function should return after the movement is finished or just started.
        """
        request = dto.LockstepStopRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
            wait_until_idle=wait_until_idle,
        )
        await call_async("device/lockstep_stop", request)

    def home(
            self,
            wait_until_idle: bool = True
    ) -> None:
        """
        Retracts the axes of the lockstep group until a home associated with an individual axis is detected.

        Args:
            wait_until_idle: Determines whether function should return after the movement is finished or just started.
        """
        request = dto.LockstepHomeRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
            wait_until_idle=wait_until_idle,
        )
        call("device/lockstep_home", request)

    async def home_async(
            self,
            wait_until_idle: bool = True
    ) -> None:
        """
        Retracts the axes of the lockstep group until a home associated with an individual axis is detected.

        Args:
            wait_until_idle: Determines whether function should return after the movement is finished or just started.
        """
        request = dto.LockstepHomeRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
            wait_until_idle=wait_until_idle,
        )
        await call_async("device/lockstep_home", request)

    def move_absolute(
            self,
            position: float,
            unit: LengthUnits = Units.NATIVE,
            wait_until_idle: bool = True,
            velocity: float = 0,
            velocity_unit: VelocityUnits = Units.NATIVE,
            acceleration: float = 0,
            acceleration_unit: AccelerationUnits = Units.NATIVE
    ) -> None:
        """
        Move the first axis of the lockstep group to an absolute position.
        The other axes in the lockstep group maintain their offsets throughout movement.

        Args:
            position: Absolute position.
            unit: Units of position.
            wait_until_idle: Determines whether function should return after the movement is finished or just started.
            velocity: Movement velocity.
                Default value of 0 indicates that the maxspeed setting is used instead.
                Requires at least Firmware 7.25.
            velocity_unit: Units of velocity.
            acceleration: Movement acceleration.
                Default value of 0 indicates that the accel setting is used instead.
                Requires at least Firmware 7.25.
            acceleration_unit: Units of acceleration.
        """
        request = dto.LockstepMoveRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
            type=dto.AxisMoveType.ABS,
            arg=position,
            unit=unit,
            wait_until_idle=wait_until_idle,
            velocity=velocity,
            velocity_unit=velocity_unit,
            acceleration=acceleration,
            acceleration_unit=acceleration_unit,
        )
        call("device/lockstep_move", request)

    async def move_absolute_async(
            self,
            position: float,
            unit: LengthUnits = Units.NATIVE,
            wait_until_idle: bool = True,
            velocity: float = 0,
            velocity_unit: VelocityUnits = Units.NATIVE,
            acceleration: float = 0,
            acceleration_unit: AccelerationUnits = Units.NATIVE
    ) -> None:
        """
        Move the first axis of the lockstep group to an absolute position.
        The other axes in the lockstep group maintain their offsets throughout movement.

        Args:
            position: Absolute position.
            unit: Units of position.
            wait_until_idle: Determines whether function should return after the movement is finished or just started.
            velocity: Movement velocity.
                Default value of 0 indicates that the maxspeed setting is used instead.
                Requires at least Firmware 7.25.
            velocity_unit: Units of velocity.
            acceleration: Movement acceleration.
                Default value of 0 indicates that the accel setting is used instead.
                Requires at least Firmware 7.25.
            acceleration_unit: Units of acceleration.
        """
        request = dto.LockstepMoveRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
            type=dto.AxisMoveType.ABS,
            arg=position,
            unit=unit,
            wait_until_idle=wait_until_idle,
            velocity=velocity,
            velocity_unit=velocity_unit,
            acceleration=acceleration,
            acceleration_unit=acceleration_unit,
        )
        await call_async("device/lockstep_move", request)

    def move_relative(
            self,
            position: float,
            unit: LengthUnits = Units.NATIVE,
            wait_until_idle: bool = True,
            velocity: float = 0,
            velocity_unit: VelocityUnits = Units.NATIVE,
            acceleration: float = 0,
            acceleration_unit: AccelerationUnits = Units.NATIVE
    ) -> None:
        """
        Move the first axis of the lockstep group to a position relative to its current position.
        The other axes in the lockstep group maintain their offsets throughout movement.

        Args:
            position: Relative position.
            unit: Units of position.
            wait_until_idle: Determines whether function should return after the movement is finished or just started.
            velocity: Movement velocity.
                Default value of 0 indicates that the maxspeed setting is used instead.
                Requires at least Firmware 7.25.
            velocity_unit: Units of velocity.
            acceleration: Movement acceleration.
                Default value of 0 indicates that the accel setting is used instead.
                Requires at least Firmware 7.25.
            acceleration_unit: Units of acceleration.
        """
        request = dto.LockstepMoveRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
            type=dto.AxisMoveType.REL,
            arg=position,
            unit=unit,
            wait_until_idle=wait_until_idle,
            velocity=velocity,
            velocity_unit=velocity_unit,
            acceleration=acceleration,
            acceleration_unit=acceleration_unit,
        )
        call("device/lockstep_move", request)

    async def move_relative_async(
            self,
            position: float,
            unit: LengthUnits = Units.NATIVE,
            wait_until_idle: bool = True,
            velocity: float = 0,
            velocity_unit: VelocityUnits = Units.NATIVE,
            acceleration: float = 0,
            acceleration_unit: AccelerationUnits = Units.NATIVE
    ) -> None:
        """
        Move the first axis of the lockstep group to a position relative to its current position.
        The other axes in the lockstep group maintain their offsets throughout movement.

        Args:
            position: Relative position.
            unit: Units of position.
            wait_until_idle: Determines whether function should return after the movement is finished or just started.
            velocity: Movement velocity.
                Default value of 0 indicates that the maxspeed setting is used instead.
                Requires at least Firmware 7.25.
            velocity_unit: Units of velocity.
            acceleration: Movement acceleration.
                Default value of 0 indicates that the accel setting is used instead.
                Requires at least Firmware 7.25.
            acceleration_unit: Units of acceleration.
        """
        request = dto.LockstepMoveRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
            type=dto.AxisMoveType.REL,
            arg=position,
            unit=unit,
            wait_until_idle=wait_until_idle,
            velocity=velocity,
            velocity_unit=velocity_unit,
            acceleration=acceleration,
            acceleration_unit=acceleration_unit,
        )
        await call_async("device/lockstep_move", request)

    def move_velocity(
            self,
            velocity: float,
            unit: VelocityUnits = Units.NATIVE,
            acceleration: float = 0,
            acceleration_unit: AccelerationUnits = Units.NATIVE
    ) -> None:
        """
        Moves the first axis of the lockstep group at the specified speed.
        The other axes in the lockstep group maintain their offsets throughout movement.

        Args:
            velocity: Movement velocity.
            unit: Units of velocity.
            acceleration: Movement acceleration.
                Default value of 0 indicates that the accel setting is used instead.
                Requires at least Firmware 7.25.
            acceleration_unit: Units of acceleration.
        """
        request = dto.LockstepMoveRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
            type=dto.AxisMoveType.VEL,
            arg=velocity,
            unit=unit,
            acceleration=acceleration,
            acceleration_unit=acceleration_unit,
        )
        call("device/lockstep_move", request)

    async def move_velocity_async(
            self,
            velocity: float,
            unit: VelocityUnits = Units.NATIVE,
            acceleration: float = 0,
            acceleration_unit: AccelerationUnits = Units.NATIVE
    ) -> None:
        """
        Moves the first axis of the lockstep group at the specified speed.
        The other axes in the lockstep group maintain their offsets throughout movement.

        Args:
            velocity: Movement velocity.
            unit: Units of velocity.
            acceleration: Movement acceleration.
                Default value of 0 indicates that the accel setting is used instead.
                Requires at least Firmware 7.25.
            acceleration_unit: Units of acceleration.
        """
        request = dto.LockstepMoveRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
            type=dto.AxisMoveType.VEL,
            arg=velocity,
            unit=unit,
            acceleration=acceleration,
            acceleration_unit=acceleration_unit,
        )
        await call_async("device/lockstep_move", request)

    def move_sin(
            self,
            amplitude: float,
            amplitude_units: LengthUnits,
            period: float,
            period_units: TimeUnits,
            count: float = 0,
            wait_until_idle: bool = True
    ) -> None:
        """
        Moves the first axis of the lockstep group in a sinusoidal trajectory.
        The other axes in the lockstep group maintain their offsets throughout movement.

        Args:
            amplitude: Amplitude of the sinusoidal motion (half of the motion's peak-to-peak range).
            amplitude_units: Units of position.
            period: Period of the sinusoidal motion in milliseconds.
            period_units: Units of time.
            count: Number of sinusoidal cycles to complete.
                Must be a multiple of 0.5
                If count is not specified or set to 0, the axis will move indefinitely.
            wait_until_idle: Determines whether function should return after the movement is finished or just started.
        """
        request = dto.LockstepMoveSinRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
            amplitude=amplitude,
            amplitude_units=amplitude_units,
            period=period,
            period_units=period_units,
            count=count,
            wait_until_idle=wait_until_idle,
        )
        call("device/lockstep_move_sin", request)

    async def move_sin_async(
            self,
            amplitude: float,
            amplitude_units: LengthUnits,
            period: float,
            period_units: TimeUnits,
            count: float = 0,
            wait_until_idle: bool = True
    ) -> None:
        """
        Moves the first axis of the lockstep group in a sinusoidal trajectory.
        The other axes in the lockstep group maintain their offsets throughout movement.

        Args:
            amplitude: Amplitude of the sinusoidal motion (half of the motion's peak-to-peak range).
            amplitude_units: Units of position.
            period: Period of the sinusoidal motion in milliseconds.
            period_units: Units of time.
            count: Number of sinusoidal cycles to complete.
                Must be a multiple of 0.5
                If count is not specified or set to 0, the axis will move indefinitely.
            wait_until_idle: Determines whether function should return after the movement is finished or just started.
        """
        request = dto.LockstepMoveSinRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
            amplitude=amplitude,
            amplitude_units=amplitude_units,
            period=period,
            period_units=period_units,
            count=count,
            wait_until_idle=wait_until_idle,
        )
        await call_async("device/lockstep_move_sin", request)

    def move_sin_stop(
            self,
            wait_until_idle: bool = True
    ) -> None:
        """
        Stops the lockstep group at the end of the sinusoidal trajectory for the first axis.
        If the sinusoidal motion was started with an integer-plus-half cycle count,
        the motion ends at the half-way point of the sinusoidal trajectory.

        Args:
            wait_until_idle: Determines whether function should return after the movement is finished.
        """
        request = dto.LockstepStopRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
            wait_until_idle=wait_until_idle,
        )
        call("device/lockstep_move_sin_stop", request)

    async def move_sin_stop_async(
            self,
            wait_until_idle: bool = True
    ) -> None:
        """
        Stops the lockstep group at the end of the sinusoidal trajectory for the first axis.
        If the sinusoidal motion was started with an integer-plus-half cycle count,
        the motion ends at the half-way point of the sinusoidal trajectory.

        Args:
            wait_until_idle: Determines whether function should return after the movement is finished.
        """
        request = dto.LockstepStopRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
            wait_until_idle=wait_until_idle,
        )
        await call_async("device/lockstep_move_sin_stop", request)

    def move_max(
            self,
            wait_until_idle: bool = True,
            velocity: float = 0,
            velocity_unit: VelocityUnits = Units.NATIVE,
            acceleration: float = 0,
            acceleration_unit: AccelerationUnits = Units.NATIVE
    ) -> None:
        """
        Moves the axes to the maximum valid position.
        The axes in the lockstep group maintain their offsets throughout movement.
        Respects lim.max for all axes in the group.

        Args:
            wait_until_idle: Determines whether function should return after the movement is finished or just started.
            velocity: Movement velocity.
                Default value of 0 indicates that the maxspeed setting is used instead.
                Requires at least Firmware 7.25.
            velocity_unit: Units of velocity.
            acceleration: Movement acceleration.
                Default value of 0 indicates that the accel setting is used instead.
                Requires at least Firmware 7.25.
            acceleration_unit: Units of acceleration.
        """
        request = dto.LockstepMoveRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
            type=dto.AxisMoveType.MAX,
            wait_until_idle=wait_until_idle,
            velocity=velocity,
            velocity_unit=velocity_unit,
            acceleration=acceleration,
            acceleration_unit=acceleration_unit,
        )
        call("device/lockstep_move", request)

    async def move_max_async(
            self,
            wait_until_idle: bool = True,
            velocity: float = 0,
            velocity_unit: VelocityUnits = Units.NATIVE,
            acceleration: float = 0,
            acceleration_unit: AccelerationUnits = Units.NATIVE
    ) -> None:
        """
        Moves the axes to the maximum valid position.
        The axes in the lockstep group maintain their offsets throughout movement.
        Respects lim.max for all axes in the group.

        Args:
            wait_until_idle: Determines whether function should return after the movement is finished or just started.
            velocity: Movement velocity.
                Default value of 0 indicates that the maxspeed setting is used instead.
                Requires at least Firmware 7.25.
            velocity_unit: Units of velocity.
            acceleration: Movement acceleration.
                Default value of 0 indicates that the accel setting is used instead.
                Requires at least Firmware 7.25.
            acceleration_unit: Units of acceleration.
        """
        request = dto.LockstepMoveRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
            type=dto.AxisMoveType.MAX,
            wait_until_idle=wait_until_idle,
            velocity=velocity,
            velocity_unit=velocity_unit,
            acceleration=acceleration,
            acceleration_unit=acceleration_unit,
        )
        await call_async("device/lockstep_move", request)

    def move_min(
            self,
            wait_until_idle: bool = True,
            velocity: float = 0,
            velocity_unit: VelocityUnits = Units.NATIVE,
            acceleration: float = 0,
            acceleration_unit: AccelerationUnits = Units.NATIVE
    ) -> None:
        """
        Moves the axes to the minimum valid position.
        The axes in the lockstep group maintain their offsets throughout movement.
        Respects lim.min for all axes in the group.

        Args:
            wait_until_idle: Determines whether function should return after the movement is finished or just started.
            velocity: Movement velocity.
                Default value of 0 indicates that the maxspeed setting is used instead.
                Requires at least Firmware 7.25.
            velocity_unit: Units of velocity.
            acceleration: Movement acceleration.
                Default value of 0 indicates that the accel setting is used instead.
                Requires at least Firmware 7.25.
            acceleration_unit: Units of acceleration.
        """
        request = dto.LockstepMoveRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
            type=dto.AxisMoveType.MIN,
            wait_until_idle=wait_until_idle,
            velocity=velocity,
            velocity_unit=velocity_unit,
            acceleration=acceleration,
            acceleration_unit=acceleration_unit,
        )
        call("device/lockstep_move", request)

    async def move_min_async(
            self,
            wait_until_idle: bool = True,
            velocity: float = 0,
            velocity_unit: VelocityUnits = Units.NATIVE,
            acceleration: float = 0,
            acceleration_unit: AccelerationUnits = Units.NATIVE
    ) -> None:
        """
        Moves the axes to the minimum valid position.
        The axes in the lockstep group maintain their offsets throughout movement.
        Respects lim.min for all axes in the group.

        Args:
            wait_until_idle: Determines whether function should return after the movement is finished or just started.
            velocity: Movement velocity.
                Default value of 0 indicates that the maxspeed setting is used instead.
                Requires at least Firmware 7.25.
            velocity_unit: Units of velocity.
            acceleration: Movement acceleration.
                Default value of 0 indicates that the accel setting is used instead.
                Requires at least Firmware 7.25.
            acceleration_unit: Units of acceleration.
        """
        request = dto.LockstepMoveRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
            type=dto.AxisMoveType.MIN,
            wait_until_idle=wait_until_idle,
            velocity=velocity,
            velocity_unit=velocity_unit,
            acceleration=acceleration,
            acceleration_unit=acceleration_unit,
        )
        await call_async("device/lockstep_move", request)

    def wait_until_idle(
            self,
            throw_error_on_fault: bool = True
    ) -> None:
        """
        Waits until the lockstep group stops moving.

        Args:
            throw_error_on_fault: Determines whether to throw error when fault is observed.
        """
        request = dto.LockstepWaitUntilIdleRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
            throw_error_on_fault=throw_error_on_fault,
        )
        call("device/lockstep_wait_until_idle", request)

    async def wait_until_idle_async(
            self,
            throw_error_on_fault: bool = True
    ) -> None:
        """
        Waits until the lockstep group stops moving.

        Args:
            throw_error_on_fault: Determines whether to throw error when fault is observed.
        """
        request = dto.LockstepWaitUntilIdleRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
            throw_error_on_fault=throw_error_on_fault,
        )
        await call_async("device/lockstep_wait_until_idle", request)

    def is_busy(
            self
    ) -> bool:
        """
        Returns bool indicating whether the lockstep group is executing a motion command.

        Returns:
            True if the axes are currently executing a motion command.
        """
        request = dto.LockstepEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
        )
        response = call(
            "device/lockstep_is_busy",
            request,
            dto.BoolResponse.from_binary)
        return response.value

    async def is_busy_async(
            self
    ) -> bool:
        """
        Returns bool indicating whether the lockstep group is executing a motion command.

        Returns:
            True if the axes are currently executing a motion command.
        """
        request = dto.LockstepEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
        )
        response = await call_async(
            "device/lockstep_is_busy",
            request,
            dto.BoolResponse.from_binary)
        return response.value

    def get_axes(
            self
    ) -> LockstepAxes:
        """
        Deprecated: Use GetAxisNumbers instead.

        Gets the axes of the lockstep group.

        Returns:
            LockstepAxes instance which contains the axes numbers of the lockstep group.
        """
        request = dto.LockstepEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
        )
        response = call(
            "device/lockstep_get_axes",
            request,
            LockstepAxes.from_binary)
        return response

    async def get_axes_async(
            self
    ) -> LockstepAxes:
        """
        Deprecated: Use GetAxisNumbers instead.

        Gets the axes of the lockstep group.

        Returns:
            LockstepAxes instance which contains the axes numbers of the lockstep group.
        """
        request = dto.LockstepEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
        )
        response = await call_async(
            "device/lockstep_get_axes",
            request,
            LockstepAxes.from_binary)
        return response

    def get_axis_numbers(
            self
    ) -> List[int]:
        """
        Gets the axis numbers of the lockstep group.

        Returns:
            Axis numbers in order specified when enabling lockstep.
        """
        request = dto.LockstepEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
        )
        response = call(
            "device/lockstep_get_axis_numbers",
            request,
            dto.LockstepGetAxisNumbersResponse.from_binary)
        return response.axes

    async def get_axis_numbers_async(
            self
    ) -> List[int]:
        """
        Gets the axis numbers of the lockstep group.

        Returns:
            Axis numbers in order specified when enabling lockstep.
        """
        request = dto.LockstepEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
        )
        response = await call_async(
            "device/lockstep_get_axis_numbers",
            request,
            dto.LockstepGetAxisNumbersResponse.from_binary)
        return response.axes

    def get_offsets(
            self,
            unit: LengthUnits = Units.NATIVE
    ) -> List[float]:
        """
        Gets the initial offsets of secondary axes of an enabled lockstep group.

        Args:
            unit: Units of position.
                Uses primary axis unit conversion.

        Returns:
            Initial offset for each axis of the lockstep group.
        """
        request = dto.LockstepGetRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
            unit=unit,
        )
        response = call(
            "device/lockstep_get_offsets",
            request,
            dto.DoubleArrayResponse.from_binary)
        return response.values

    async def get_offsets_async(
            self,
            unit: LengthUnits = Units.NATIVE
    ) -> List[float]:
        """
        Gets the initial offsets of secondary axes of an enabled lockstep group.

        Args:
            unit: Units of position.
                Uses primary axis unit conversion.

        Returns:
            Initial offset for each axis of the lockstep group.
        """
        request = dto.LockstepGetRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
            unit=unit,
        )
        response = await call_async(
            "device/lockstep_get_offsets",
            request,
            dto.DoubleArrayResponse.from_binary)
        return response.values

    def get_twists(
            self,
            unit: LengthUnits = Units.NATIVE
    ) -> List[float]:
        """
        Gets the twists of secondary axes of an enabled lockstep group.

        Args:
            unit: Units of position.
                Uses primary axis unit conversion.

        Returns:
            Difference between the initial offset and the actual offset for each axis of the lockstep group.
        """
        request = dto.LockstepGetRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
            unit=unit,
        )
        response = call(
            "device/lockstep_get_twists",
            request,
            dto.DoubleArrayResponse.from_binary)
        return response.values

    async def get_twists_async(
            self,
            unit: LengthUnits = Units.NATIVE
    ) -> List[float]:
        """
        Gets the twists of secondary axes of an enabled lockstep group.

        Args:
            unit: Units of position.
                Uses primary axis unit conversion.

        Returns:
            Difference between the initial offset and the actual offset for each axis of the lockstep group.
        """
        request = dto.LockstepGetRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
            unit=unit,
        )
        response = await call_async(
            "device/lockstep_get_twists",
            request,
            dto.DoubleArrayResponse.from_binary)
        return response.values

    def get_position(
            self,
            unit: LengthUnits = Units.NATIVE
    ) -> float:
        """
        Returns current position of the primary axis.

        Args:
            unit: Units of the position.

        Returns:
            Primary axis position.
        """
        request = dto.LockstepGetRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
            unit=unit,
        )
        response = call(
            "device/lockstep_get_pos",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    async def get_position_async(
            self,
            unit: LengthUnits = Units.NATIVE
    ) -> float:
        """
        Returns current position of the primary axis.

        Args:
            unit: Units of the position.

        Returns:
            Primary axis position.
        """
        request = dto.LockstepGetRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
            unit=unit,
        )
        response = await call_async(
            "device/lockstep_get_pos",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    def set_tolerance(
            self,
            tolerance: float,
            unit: LengthUnits = Units.NATIVE,
            axis_index: int = 0
    ) -> None:
        """
        Sets lockstep twist tolerance.
        Twist tolerances that do not match the system configuration can reduce performance or damage the system.

        Args:
            tolerance: Twist tolerance.
            unit: Units of the tolerance.
                Uses primary axis unit conversion when setting to all axes,
                otherwise uses specified secondary axis unit conversion.
            axis_index: Optional index of a secondary axis to set the tolerance for.
                If left empty or set to 0, the tolerance is set to all the secondary axes.
        """
        request = dto.LockstepSetRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
            value=tolerance,
            unit=unit,
            axis_index=axis_index,
        )
        call("device/lockstep_set_tolerance", request)

    async def set_tolerance_async(
            self,
            tolerance: float,
            unit: LengthUnits = Units.NATIVE,
            axis_index: int = 0
    ) -> None:
        """
        Sets lockstep twist tolerance.
        Twist tolerances that do not match the system configuration can reduce performance or damage the system.

        Args:
            tolerance: Twist tolerance.
            unit: Units of the tolerance.
                Uses primary axis unit conversion when setting to all axes,
                otherwise uses specified secondary axis unit conversion.
            axis_index: Optional index of a secondary axis to set the tolerance for.
                If left empty or set to 0, the tolerance is set to all the secondary axes.
        """
        request = dto.LockstepSetRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
            value=tolerance,
            unit=unit,
            axis_index=axis_index,
        )
        await call_async("device/lockstep_set_tolerance", request)

    def is_enabled(
            self
    ) -> bool:
        """
        Checks if the lockstep group is currently enabled on the device.

        Returns:
            True if a lockstep group with this ID is enabled on the device.
        """
        request = dto.LockstepEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
        )
        response = call(
            "device/lockstep_is_enabled",
            request,
            dto.BoolResponse.from_binary)
        return response.value

    async def is_enabled_async(
            self
    ) -> bool:
        """
        Checks if the lockstep group is currently enabled on the device.

        Returns:
            True if a lockstep group with this ID is enabled on the device.
        """
        request = dto.LockstepEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
        )
        response = await call_async(
            "device/lockstep_is_enabled",
            request,
            dto.BoolResponse.from_binary)
        return response.value

    def __repr__(
            self
    ) -> str:
        """
        Returns a string which represents the enabled lockstep group.

        Returns:
            String which represents the enabled lockstep group.
        """
        request = dto.LockstepEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            lockstep_group_id=self.lockstep_group_id,
        )
        response = call_sync(
            "device/lockstep_to_string",
            request,
            dto.StringResponse.from_binary)
        return response.value



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long
from .autofocus_status import AutofocusStatus as AutofocusStatus
from .microscope_config import MicroscopeConfig as MicroscopeConfig
from .third_party_components import ThirdPartyComponents as ThirdPartyComponents
from .wdi_autofocus_provider_status import WdiAutofocusProviderStatus as WdiAutofocusProviderStatus



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #

from typing import TYPE_CHECKING
from ..call import call, call_async, call_sync

from ..dto import requests as dto

if TYPE_CHECKING:
    from .device import Device


class AllAxes:
    """
    Represents all axes of motion associated with a device.
    """

    @property
    def device(self) -> 'Device':
        """
        Device that controls this axis.
        """
        return self._device

    def __init__(self, device: 'Device'):
        self._device = device

    def home(
            self,
            wait_until_idle: bool = True
    ) -> None:
        """
        Homes all axes. Axes return to their homing positions.

        Args:
            wait_until_idle: Determines whether function should return after the movement is finished or just started.
        """
        request = dto.DeviceHomeRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            axis=0,
            wait_until_idle=wait_until_idle,
        )
        call("device/home", request)

    async def home_async(
            self,
            wait_until_idle: bool = True
    ) -> None:
        """
        Homes all axes. Axes return to their homing positions.

        Args:
            wait_until_idle: Determines whether function should return after the movement is finished or just started.
        """
        request = dto.DeviceHomeRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            axis=0,
            wait_until_idle=wait_until_idle,
        )
        await call_async("device/home", request)

    def stop(
            self,
            wait_until_idle: bool = True
    ) -> None:
        """
        Stops ongoing axes movement. Decelerates until zero speed.

        Args:
            wait_until_idle: Determines whether function should return after the movement is finished or just started.
        """
        request = dto.DeviceStopRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            axis=0,
            wait_until_idle=wait_until_idle,
        )
        call("device/stop", request)

    async def stop_async(
            self,
            wait_until_idle: bool = True
    ) -> None:
        """
        Stops ongoing axes movement. Decelerates until zero speed.

        Args:
            wait_until_idle: Determines whether function should return after the movement is finished or just started.
        """
        request = dto.DeviceStopRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            axis=0,
            wait_until_idle=wait_until_idle,
        )
        await call_async("device/stop", request)

    def wait_until_idle(
            self,
            throw_error_on_fault: bool = True
    ) -> None:
        """
        Waits until all axes of device stop moving.

        Args:
            throw_error_on_fault: Determines whether to throw error when fault is observed.
        """
        request = dto.DeviceWaitUntilIdleRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            axis=0,
            throw_error_on_fault=throw_error_on_fault,
        )
        call("device/wait_until_idle", request)

    async def wait_until_idle_async(
            self,
            throw_error_on_fault: bool = True
    ) -> None:
        """
        Waits until all axes of device stop moving.

        Args:
            throw_error_on_fault: Determines whether to throw error when fault is observed.
        """
        request = dto.DeviceWaitUntilIdleRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            axis=0,
            throw_error_on_fault=throw_error_on_fault,
        )
        await call_async("device/wait_until_idle", request)

    def park(
            self
    ) -> None:
        """
        Parks the device in anticipation of turning the power off.
        It can later be powered on, unparked, and moved without first having to home it.
        """
        request = dto.AxisEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            axis=0,
        )
        call("device/park", request)

    async def park_async(
            self
    ) -> None:
        """
        Parks the device in anticipation of turning the power off.
        It can later be powered on, unparked, and moved without first having to home it.
        """
        request = dto.AxisEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            axis=0,
        )
        await call_async("device/park", request)

    def unpark(
            self
    ) -> None:
        """
        Unparks the device. The device will now be able to move.
        """
        request = dto.AxisEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            axis=0,
        )
        call("device/unpark", request)

    async def unpark_async(
            self
    ) -> None:
        """
        Unparks the device. The device will now be able to move.
        """
        request = dto.AxisEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            axis=0,
        )
        await call_async("device/unpark", request)

    def is_busy(
            self
    ) -> bool:
        """
        Returns bool indicating whether any axis is executing a motion command.

        Returns:
            True if any axis is currently executing a motion command.
        """
        request = dto.AxisEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            axis=0,
        )
        response = call(
            "device/is_busy",
            request,
            dto.BoolResponse.from_binary)
        return response.value

    async def is_busy_async(
            self
    ) -> bool:
        """
        Returns bool indicating whether any axis is executing a motion command.

        Returns:
            True if any axis is currently executing a motion command.
        """
        request = dto.AxisEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            axis=0,
        )
        response = await call_async(
            "device/is_busy",
            request,
            dto.BoolResponse.from_binary)
        return response.value

    def is_homed(
            self
    ) -> bool:
        """
        Returns bool indicating whether all axes have position reference and were homed.

        Returns:
            True if all axes have position reference and were homed.
        """
        request = dto.AxisEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            axis=0,
        )
        response = call(
            "device/is_homed",
            request,
            dto.BoolResponse.from_binary)
        return response.value

    async def is_homed_async(
            self
    ) -> bool:
        """
        Returns bool indicating whether all axes have position reference and were homed.

        Returns:
            True if all axes have position reference and were homed.
        """
        request = dto.AxisEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            axis=0,
        )
        response = await call_async(
            "device/is_homed",
            request,
            dto.BoolResponse.from_binary)
        return response.value

    def driver_disable(
            self
    ) -> None:
        """
        Disables all axes drivers, which prevents current from being sent to the motor or load.
        """
        request = dto.AxisEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            axis=0,
        )
        call("device/driver_disable", request)

    async def driver_disable_async(
            self
    ) -> None:
        """
        Disables all axes drivers, which prevents current from being sent to the motor or load.
        """
        request = dto.AxisEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            axis=0,
        )
        await call_async("device/driver_disable", request)

    def driver_enable(
            self,
            timeout: float = 10
    ) -> None:
        """
        Attempts to enable all axes drivers (where applicable) repeatedly for the specified timeout.
        If the driver is already enabled, the driver remains enabled.

        Args:
            timeout: Timeout in seconds. Specify 0 to attempt to enable the driver once.
        """
        request = dto.DriverEnableRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            axis=0,
            timeout=timeout,
        )
        call("device/driver_enable", request)

    async def driver_enable_async(
            self,
            timeout: float = 10
    ) -> None:
        """
        Attempts to enable all axes drivers (where applicable) repeatedly for the specified timeout.
        If the driver is already enabled, the driver remains enabled.

        Args:
            timeout: Timeout in seconds. Specify 0 to attempt to enable the driver once.
        """
        request = dto.DriverEnableRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            axis=0,
            timeout=timeout,
        )
        await call_async("device/driver_enable", request)

    def __repr__(
            self
    ) -> str:
        """
        Returns a string that represents the axes.

        Returns:
            A string that represents the axes.
        """
        request = dto.AxisEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            axis=0,
        )
        response = call_sync(
            "device/all_axes_to_string",
            request,
            dto.StringResponse.from_binary)
        return response.value



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #
# pylint: disable=W0622

from ..call import call, call_async, call_sync
from ..dto import requests as dto
from ..ascii import Device


class FilterChanger:
    """
    A generic turret device.
    """

    @property
    def device(self) -> Device:
        """
        The base device of this turret.
        """
        return self._device

    def __init__(self, device: Device):
        """
        Creates instance of `FilterChanger` based on the given device.
        """
        self._device = device

    def get_number_of_filters(
            self
    ) -> int:
        """
        Gets number of filters of the changer.

        Returns:
            Number of positions.
        """
        request = dto.AxisEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            axis=1,
        )
        response = call(
            "device/get_index_count",
            request,
            dto.IntResponse.from_binary)
        return response.value

    async def get_number_of_filters_async(
            self
    ) -> int:
        """
        Gets number of filters of the changer.

        Returns:
            Number of positions.
        """
        request = dto.AxisEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            axis=1,
        )
        response = await call_async(
            "device/get_index_count",
            request,
            dto.IntResponse.from_binary)
        return response.value

    def get_current_filter(
            self
    ) -> int:
        """
        Returns the current filter number starting from 1.
        The value of 0 indicates that the position is either unknown or between two filters.

        Returns:
            Filter number starting from 1 or 0 if the position cannot be determined.
        """
        request = dto.AxisEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            axis=1,
        )
        response = call(
            "device/get_index_position",
            request,
            dto.IntResponse.from_binary)
        return response.value

    async def get_current_filter_async(
            self
    ) -> int:
        """
        Returns the current filter number starting from 1.
        The value of 0 indicates that the position is either unknown or between two filters.

        Returns:
            Filter number starting from 1 or 0 if the position cannot be determined.
        """
        request = dto.AxisEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            axis=1,
        )
        response = await call_async(
            "device/get_index_position",
            request,
            dto.IntResponse.from_binary)
        return response.value

    def change(
            self,
            filter: int
    ) -> None:
        """
        Changes to the specified filter.

        Args:
            filter: Filter number starting from 1.
        """
        request = dto.DeviceMoveRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            axis=1,
            type=dto.AxisMoveType.INDEX,
            wait_until_idle=True,
            arg_int=filter,
        )
        call("device/move", request)

    async def change_async(
            self,
            filter: int
    ) -> None:
        """
        Changes to the specified filter.

        Args:
            filter: Filter number starting from 1.
        """
        request = dto.DeviceMoveRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            axis=1,
            type=dto.AxisMoveType.INDEX,
            wait_until_idle=True,
            arg_int=filter,
        )
        await call_async("device/move", request)

    def __repr__(
            self
    ) -> str:
        """
        Returns a string that represents the device.

        Returns:
            A string that represents the device.
        """
        request = dto.AxisToStringRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
        )
        response = call_sync(
            "device/device_to_string",
            request,
            dto.StringResponse.from_binary)
        return response.value



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict
import decimal
import zaber_bson


@dataclass
class GenericCommandRequest:

    interface_id: int = 0

    device: int = 0

    axis: int = 0

    command: str = ""

    check_errors: bool = False

    timeout: int = 0

    @staticmethod
    def zero_values() -> 'GenericCommandRequest':
        return GenericCommandRequest(
            interface_id=0,
            device=0,
            axis=0,
            command="",
            check_errors=False,
            timeout=0,
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'GenericCommandRequest':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return GenericCommandRequest.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'interfaceId': int(self.interface_id),
            'device': int(self.device),
            'axis': int(self.axis),
            'command': str(self.command or ''),
            'checkErrors': bool(self.check_errors),
            'timeout': int(self.timeout),
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'GenericCommandRequest':
        return GenericCommandRequest(
            interface_id=data.get('interfaceId'),  # type: ignore
            device=data.get('device'),  # type: ignore
            axis=data.get('axis'),  # type: ignore
            command=data.get('command'),  # type: ignore
            check_errors=data.get('checkErrors'),  # type: ignore
            timeout=data.get('timeout'),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.interface_id is None:
            raise ValueError(f'Property "InterfaceId" of "GenericCommandRequest" is None.')

        if not isinstance(self.interface_id, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "InterfaceId" of "GenericCommandRequest" is not a number.')

        if int(self.interface_id) != self.interface_id:
            raise ValueError(f'Property "InterfaceId" of "GenericCommandRequest" is not integer value.')

        if self.device is None:
            raise ValueError(f'Property "Device" of "GenericCommandRequest" is None.')

        if not isinstance(self.device, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Device" of "GenericCommandRequest" is not a number.')

        if int(self.device) != self.device:
            raise ValueError(f'Property "Device" of "GenericCommandRequest" is not integer value.')

        if self.axis is None:
            raise ValueError(f'Property "Axis" of "GenericCommandRequest" is None.')

        if not isinstance(self.axis, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Axis" of "GenericCommandRequest" is not a number.')

        if int(self.axis) != self.axis:
            raise ValueError(f'Property "Axis" of "GenericCommandRequest" is not integer value.')

        if self.command is not None:
            if not isinstance(self.command, str):
                raise ValueError(f'Property "Command" of "GenericCommandRequest" is not a string.')

        if self.timeout is None:
            raise ValueError(f'Property "Timeout" of "GenericCommandRequest" is None.')

        if not isinstance(self.timeout, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Timeout" of "GenericCommandRequest" is not a number.')

        if int(self.timeout) != self.timeout:
            raise ValueError(f'Property "Timeout" of "GenericCommandRequest" is not integer value.')



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass, field
from typing import Any, Dict, List
import decimal
from collections.abc import Iterable
import zaber_bson


@dataclass
class WdiGenericRequest:

    interface_id: int = 0

    register_id: int = 0

    size: int = 0

    count: int = 0

    offset: int = 0

    register_bank: str = ""

    data: List[int] = field(default_factory=list)

    @staticmethod
    def zero_values() -> 'WdiGenericRequest':
        return WdiGenericRequest(
            interface_id=0,
            register_id=0,
            size=0,
            count=0,
            offset=0,
            register_bank="",
            data=[],
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'WdiGenericRequest':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return WdiGenericRequest.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'interfaceId': int(self.interface_id),
            'registerId': int(self.register_id),
            'size': int(self.size),
            'count': int(self.count),
            'offset': int(self.offset),
            'registerBank': str(self.register_bank or ''),
            'data': [int(item) for item in self.data] if self.data is not None else [],
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'WdiGenericRequest':
        return WdiGenericRequest(
            interface_id=data.get('interfaceId'),  # type: ignore
            register_id=data.get('registerId'),  # type: ignore
            size=data.get('size'),  # type: ignore
            count=data.get('count'),  # type: ignore
            offset=data.get('offset'),  # type: ignore
            register_bank=data.get('registerBank'),  # type: ignore
            data=data.get('data'),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.interface_id is None:
            raise ValueError(f'Property "InterfaceId" of "WdiGenericRequest" is None.')

        if not isinstance(self.interface_id, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "InterfaceId" of "WdiGenericRequest" is not a number.')

        if int(self.interface_id) != self.interface_id:
            raise ValueError(f'Property "InterfaceId" of "WdiGenericRequest" is not integer value.')

        if self.register_id is None:
            raise ValueError(f'Property "RegisterId" of "WdiGenericRequest" is None.')

        if not isinstance(self.register_id, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "RegisterId" of "WdiGenericRequest" is not a number.')

        if int(self.register_id) != self.register_id:
            raise ValueError(f'Property "RegisterId" of "WdiGenericRequest" is not integer value.')

        if self.size is None:
            raise ValueError(f'Property "Size" of "WdiGenericRequest" is None.')

        if not isinstance(self.size, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Size" of "WdiGenericRequest" is not a number.')

        if int(self.size) != self.size:
            raise ValueError(f'Property "Size" of "WdiGenericRequest" is not integer value.')

        if self.count is None:
            raise ValueError(f'Property "Count" of "WdiGenericRequest" is None.')

        if not isinstance(self.count, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Count" of "WdiGenericRequest" is not a number.')

        if int(self.count) != self.count:
            raise ValueError(f'Property "Count" of "WdiGenericRequest" is not integer value.')

        if self.offset is None:
            raise ValueError(f'Property "Offset" of "WdiGenericRequest" is None.')

        if not isinstance(self.offset, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Offset" of "WdiGenericRequest" is not a number.')

        if int(self.offset) != self.offset:
            raise ValueError(f'Property "Offset" of "WdiGenericRequest" is not integer value.')

        if self.register_bank is not None:
            if not isinstance(self.register_bank, str):
                raise ValueError(f'Property "RegisterBank" of "WdiGenericRequest" is not a string.')

        if self.data is not None:
            if not isinstance(self.data, Iterable):
                raise ValueError('Property "Data" of "WdiGenericRequest" is not iterable.')

            for i, data_item in enumerate(self.data):
                if data_item is None:
                    raise ValueError(f'Item {i} in property "Data" of "WdiGenericRequest" is None.')

                if not isinstance(data_item, (int, float, decimal.Decimal)):
                    raise ValueError(f'Item {i} in property "Data" of "WdiGenericRequest" is not a number.')

                if int(data_item) != data_item:
                    raise ValueError(f'Item {i} in property "Data" of "WdiGenericRequest" is not integer value.')



import os
import re
import urllib.request
import shutil
import setuptools
import zipfile


PACKAGE_DIR = os.path.dirname(__file__)
BINDINGS_DIR = os.path.join(PACKAGE_DIR, "zaber_motion_bindings")


def get_version():
    with open(os.path.join(PACKAGE_DIR, "pyproject.toml"), "r", encoding="utf-8") as fh:
        pyproject = fh.read()

    return re.search(r'version = "(.*?)"', pyproject).group(1)


def download_bindings(version):
    url = f"https://api.zaber.io/downloads/software-downloads?product=zml_cpp_src&version={version}"
    filename = f"ZaberMotionCppSource-{version}.zip"
    bindings_prefix = "ZaberMotionCppSource/build/zaber-motion-lib"

    with urllib.request.urlopen(url) as response, open(filename, 'wb') as out_file:
        shutil.copyfileobj(response, out_file)

    with zipfile.ZipFile(filename, 'r') as archive:
        files = archive.namelist()
        bindings = [f for f in files if f.startswith(bindings_prefix)]
        archive.extractall(".", bindings)

    shutil.rmtree(BINDINGS_DIR, ignore_errors=True)
    shutil.move("ZaberMotionCppSource/build", BINDINGS_DIR)
    shutil.rmtree("ZaberMotionCppSource", ignore_errors=True)
    os.remove(filename)


def setup():
    if not os.environ.get("SETUP_SKIP_ZML_BINDINGS"):
        version = get_version()
        download_bindings(version)

    setuptools.setup()


setup()



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass, field
from typing import Any, Dict
import zaber_bson
from ..microscopy.autofocus_status import AutofocusStatus


@dataclass
class AutofocusGetStatusResponse:

    status: AutofocusStatus = field(default_factory=AutofocusStatus.zero_values)

    @staticmethod
    def zero_values() -> 'AutofocusGetStatusResponse':
        return AutofocusGetStatusResponse(
            status=AutofocusStatus.zero_values(),
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'AutofocusGetStatusResponse':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return AutofocusGetStatusResponse.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'status': self.status.to_dict(),
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'AutofocusGetStatusResponse':
        return AutofocusGetStatusResponse(
            status=AutofocusStatus.from_dict(data.get('status')),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.status is None:
            raise ValueError(f'Property "Status" of "AutofocusGetStatusResponse" is None.')

        if not isinstance(self.status, AutofocusStatus):
            raise ValueError(f'Property "Status" of "AutofocusGetStatusResponse" is not an instance of "AutofocusStatus".')

        self.status.validate()



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict, Optional
import decimal
import zaber_bson


@dataclass
class AxisDefinition:
    """
    Defines an axis of the translator.
    """

    peripheral_id: int
    """
    ID of the peripheral.
    """

    microstep_resolution: Optional[int] = None
    """
    Microstep resolution of the axis.
    Can be obtained by reading the resolution setting.
    Leave empty if the axis does not have the setting.
    """

    @staticmethod
    def zero_values() -> 'AxisDefinition':
        return AxisDefinition(
            peripheral_id=0,
            microstep_resolution=None,
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'AxisDefinition':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return AxisDefinition.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'peripheralId': int(self.peripheral_id),
            'microstepResolution': int(self.microstep_resolution) if self.microstep_resolution is not None else None,
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'AxisDefinition':
        return AxisDefinition(
            peripheral_id=data.get('peripheralId'),  # type: ignore
            microstep_resolution=data.get('microstepResolution'),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.peripheral_id is None:
            raise ValueError(f'Property "PeripheralId" of "AxisDefinition" is None.')

        if not isinstance(self.peripheral_id, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "PeripheralId" of "AxisDefinition" is not a number.')

        if int(self.peripheral_id) != self.peripheral_id:
            raise ValueError(f'Property "PeripheralId" of "AxisDefinition" is not integer value.')

        if self.microstep_resolution is not None:
            if not isinstance(self.microstep_resolution, (int, float, decimal.Decimal)):
                raise ValueError(f'Property "MicrostepResolution" of "AxisDefinition" is not a number.')

            if int(self.microstep_resolution) != self.microstep_resolution:
                raise ValueError(f'Property "MicrostepResolution" of "AxisDefinition" is not integer value.')



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #

from typing import Optional
from .call import call_sync
from .dto import requests as dto
from .dto.log_output_mode import LogOutputMode
from .dto.device_db_source_type import DeviceDbSourceType


class Library:
    """
    Access class to general library information and configuration.
    """

    @staticmethod
    def set_log_output(
            mode: LogOutputMode,
            file_path: Optional[str] = None
    ) -> None:
        """
        Sets library logging output.

        Args:
            mode: Logging output mode.
            file_path: Path of the file to open.
        """
        request = dto.SetLogOutputRequest(
            mode=mode,
            file_path=file_path,
        )
        call_sync("logging/set_output", request)

    @staticmethod
    def set_device_db_source(
            source_type: DeviceDbSourceType,
            url_or_file_path: Optional[str] = None
    ) -> None:
        """
        Sets source of Device DB data. Allows selection of a web service or a local file.

        Args:
            source_type: Source type.
            url_or_file_path: URL of the web service or path to the local file.
                Leave empty for the default URL of Zaber web service.
        """
        request = dto.SetDeviceDbSourceRequest(
            source_type=source_type,
            url_or_file_path=url_or_file_path,
        )
        call_sync("device_db/set_source", request)

    @staticmethod
    def enable_device_db_store(
            store_location: Optional[str] = None
    ) -> None:
        """
        Enables Device DB store.
        The store uses filesystem to save information obtained from the Device DB.
        The stored data are later used instead of the Device DB.

        Args:
            store_location: Specifies relative or absolute path of the folder used by the store.
                If left empty defaults to a folder in user home directory.
                Must be accessible by the process.
        """
        request = dto.ToggleDeviceDbStoreRequest(
            toggle_on=True,
            store_location=store_location,
        )
        call_sync("device_db/toggle_store", request)

    @staticmethod
    def disable_device_db_store() -> None:
        """
        Disables Device DB store.
        """
        request = dto.ToggleDeviceDbStoreRequest(
        )
        call_sync("device_db/toggle_store", request)

    @staticmethod
    def set_internal_mode(
            mode: bool
    ) -> None:
        """
        Disables certain customer checks (like FF flag).

        Args:
            mode: Whether to turn internal mode on or off.
        """
        request = dto.SetInternalModeRequest(
            mode=mode,
        )
        call_sync("library/set_internal_mode", request)

    @staticmethod
    def set_idle_polling_period(
            period: int
    ) -> None:
        """
        Sets the period between polling for IDLE during movements.
        Caution: Setting the period too low may cause performance issues.

        Args:
            period: Period in milliseconds.
                Negative value restores the default period.
        """
        request = dto.IntRequest(
            value=period,
        )
        call_sync("library/set_idle_polling_period", request)

    @staticmethod
    def check_version() -> None:
        """
        Throws an error if the version of the loaded shared library does not match the caller's version.
        """
        request = dto.CheckVersionRequest(
            host="py",
            version="7.4.0",
        )
        call_sync("library/check_version", request)


Library.check_version()



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #

from typing import Union
from ..dto.exceptions.movement_failed_exception_data import MovementFailedExceptionData
from .motion_lib_exception import MotionLibException


class MovementFailedException(MotionLibException):
    """
    Thrown when a device registers a fault during movement.
    """

    @property
    def details(self) -> MovementFailedExceptionData:
        """
        Additional data for MovementFailedException
        """
        return self._details

    def __init__(self, message: str, custom_data: Union[bytes, MovementFailedExceptionData]):
        MotionLibException.__init__(self, message)

        if isinstance(custom_data, MovementFailedExceptionData):
            self._details = custom_data
        else:
            self._details = MovementFailedExceptionData.from_binary(custom_data)
